@@ -10,7 +10,9 @@ class Command(BaseGenerateCommand):
     help = "Generate scaffolding to compile javascript."
 
     def handle(self, *args, **options):
+        init = 'npm init -y'
         install = 'npm install -g add-project-script'
+        subprocess.check_call(init, shell=True)
         subprocess.check_call(install, shell=True)
         try:
             build = 'add-project-script -n "build" -v "webpack --mode production"'
@@ -23,7 +25,7 @@ class Command(BaseGenerateCommand):
 
         subprocess.check_call('npm uninstall -g add-project-script', shell=True)
 
-        npm_pkg = 'npm install -save-dev glob sockpuppet-js stimulus_reflex webpack webpack-cli'
+        npm_pkg = 'npm install -save-dev glob sockpuppet-js stimulus stimulus_reflex webpack webpack-cli'
         subprocess.check_call(npm_pkg.split(' '))
 
         template = get_template('sockpuppet/scaffolds/webpack.html')
